/* 
 * File:   mzload.c
 * Author: Michal Hucik <hucik@ordoz.com>
 *
 * Created on 25. září 2018, 10:26
 * 
 * 
 * ----------------------------- License -------------------------------------
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * ---------------------------------------------------------------------------
 */

#include <glib.h>
#include <string.h>

#include "libs/generic_driver/generic_driver.h"
#include "libs/cmt_stream/cmt_stream.h"
#include "ui/ui_utils.h"
#include "ui/generic_driver/ui_memory_driver.h"

#include "libs/mzf/mzf.h"
#include "libs/mzf/mzf_tools.h"

#include "mzload_turbo.h"
#include "mzload_fastipl.h"


static st_DRIVER *g_driver = &g_ui_memory_driver_static;

#define MZLOAD_PILOT_MIN_PULSES    4000

#define MZLOAD_READ_SYNC    256 /* pocet prvnich pulzu, ktere nechame projit bez povsimnuti */

#define MZLOAD_SYNCHRO_BLOCK_SIZE 0x0200
#define MZLOAD_SYNCHRO_BLOCK_DIVIDER 0x0180


typedef enum en_MZLOAD_TAPEMARK {
    MZLOAD_TAPEMARK_HEADER = 40,
    MZLOAD_TAPEMARK_BODY = 20,
} en_MZLOAD_TAPEMARK;


typedef enum en_MZLOAD_STATUS {
    MZLOAD_STATUS_OK = 0,
    MZLOAD_STATUS_EOF,
    MZLOAD_STATUS_ERR_TMARK,
    MZLOAD_STATUS_ERR_DATA,
    MZLOAD_STATUS_ERR_CHECKSUM,
} en_MZLOAD_STATUS;


typedef struct st_MZLOAD_FILE {
    st_MZF_HEADER *hdr;
    uint8_t *data;
} st_MZLOAD_FILE;

#define MZLOAD_PILOT_SYNC_SAMPLES 32
#define MZLOAD_PILOT_SYNC_ALLOWED_DIFERENCE 2


typedef struct st_MZLOAD_PILOT {
    uint32_t last0_length[MZLOAD_PILOT_SYNC_SAMPLES];
    uint32_t last1_length[MZLOAD_PILOT_SYNC_SAMPLES];
    uint32_t length0_min;
    uint32_t length0_max;
    uint32_t length1_min;
    uint32_t length1_max;
    int sync_pos;
    uint32_t same_pulses;
    gboolean minmax_synchronised;
    gboolean have_pilot_start;
    gboolean have_pilot_end;
    uint32_t start_pos;
    uint32_t end_pos;
} st_MZLOAD_PILOT;


typedef struct st_MZLOAD {
    st_CMT_STREAM *stream;
    uint32_t pos; /* pocet vzorku - pozice ve streamu */
    int32_t sampTm; /* pocet vzorku - aktualni sample time (readpoint) */
    int32_t loadTm; /* delka posledniho cteneho pulzu */
    int bit_value;
    st_MZLOAD_PILOT pilot;
    en_MZLOAD_STATUS status;
} st_MZLOAD;


void mzload_destroy ( st_MZLOAD *mzload ) {
    if ( !mzload ) return;
    if ( mzload->stream ) cmt_stream_destroy ( mzload->stream );
    //if ( mzload->current_mzf ) ui_utils_mem_free ( mzload->current_mzf );
    /*
    if ( mzload->mzf_count ) {
        int i;
        for ( i = 0; i < mzload->mzf_count; i++ ) {
            ui_utils_mem_free ( mzload->mzfs[i] );
        };
        ui_utils_mem_free ( mzload->mzfs );
    };
     */
    ui_utils_mem_free ( mzload );
}


static void mzload_reset_pilot ( st_MZLOAD *mzload ) {
    st_MZLOAD_PILOT *pilot = &mzload->pilot;
    memset ( pilot, 0x00, sizeof ( st_MZLOAD_PILOT ) );
    pilot->minmax_synchronised = FALSE;
    pilot->have_pilot_start = FALSE;
    pilot->have_pilot_end = FALSE;
}


static void mzload_pilot_reset_minmax_sync ( st_MZLOAD *mzload ) {
    st_MZLOAD_PILOT *pilot = &mzload->pilot;
    memset ( pilot->last0_length, 0x00, MZLOAD_PILOT_SYNC_SAMPLES );
    memset ( pilot->last1_length, 0x00, MZLOAD_PILOT_SYNC_SAMPLES );
    pilot->minmax_synchronised = FALSE;
}


st_MZLOAD *mzload_create ( char *filename ) {

    st_HANDLER *h = generic_driver_open_memory_from_file ( NULL, g_driver, filename );
    if ( !h ) {
        fprintf ( stderr, "%s():%d - Can't open file '%s'\n", __func__, __LINE__, filename );
        return NULL;
    };

    st_MZLOAD *mzload = (st_MZLOAD*) ui_utils_mem_alloc0 ( sizeof ( st_MZLOAD ) );
    if ( !mzload ) {
        fprintf ( stderr, "%s():%d - Can't alocate memory (%d)\n", __func__, __LINE__, (int) sizeof ( st_MZLOAD ) );
        return NULL;
    };

    mzload->stream = cmt_stream_new_from_wav ( h, CMT_STREAM_POLARITY_NORMAL, CMT_STREAM_TYPE_BITSTREAM );
    generic_driver_close ( h );
    if ( !mzload->stream ) {
        mzload_destroy ( mzload );
        return NULL;
    };

    mzload_reset_pilot ( mzload );

    generic_driver_close ( h );

    return mzload;
}


static inline gboolean mzload_pilot_minmax ( uint32_t *length, int count, uint32_t *min, uint32_t *max, uint32_t *sum ) {
    *max = 0;
    *min = 0xffffffff;
    *sum = 0;
    int i;
    for ( i = 0; i < count; i++ ) {
        if ( length[i] <= 3 ) return FALSE;
        *sum = *sum + length[i];
        if ( length[i] < *min ) *min = length[i];
        if ( length[i] > *max ) *max = length[i];
    };
    if ( ( *max - *min ) <= MZLOAD_PILOT_SYNC_ALLOWED_DIFERENCE ) return TRUE;
    return FALSE;
}


static inline void mzload_pilot_add_pulses ( st_MZLOAD *mzload, uint32_t length0, uint32_t length1 ) {
    st_MZLOAD_PILOT *pilot = &mzload->pilot;
    int pos = pilot->sync_pos & ( MZLOAD_PILOT_SYNC_SAMPLES - 1 );
    //printf ( "p0: %d, p1: %d (%d)\n", length0, length1, pilot->same_pulses );
    pilot->last0_length[pos] = length0;
    pilot->last1_length[pos] = length1;
    pilot->sync_pos++;

    uint32_t sum0 = 0;
    uint32_t sum1 = 1;

    gboolean sync0 = mzload_pilot_minmax ( pilot->last0_length, MZLOAD_PILOT_SYNC_SAMPLES, &pilot->length0_min, &pilot->length0_max, &sum0 );
    gboolean sync1 = mzload_pilot_minmax ( pilot->last1_length, MZLOAD_PILOT_SYNC_SAMPLES, &pilot->length1_min, &pilot->length1_max, &sum1 );

    if ( sync0 && sync1 ) {
        pilot->minmax_synchronised = TRUE;
        pilot->start_pos = mzload->pos - ( sum0 + sum1 );
    };
}


static inline gboolean mzload_pilot_search_start ( st_MZLOAD *mzload, uint32_t length0, uint32_t length1 ) {
    st_MZLOAD_PILOT *pilot = &mzload->pilot;
    if ( ( length0 >= ( pilot->length0_min - 1 ) ) && ( length0 <= ( pilot->length0_max + 1 ) ) ) {
        if ( ( length1 >= ( pilot->length1_min - 1 ) ) && ( length1 <= ( pilot->length1_max + 1 ) ) ) {
            pilot->same_pulses++;
            if ( pilot->same_pulses >= MZLOAD_PILOT_MIN_PULSES ) {
                pilot->have_pilot_start = TRUE;
                printf ( "Pilot start pos: %d, length0: <%d, %d>, length1: <%d, %d>\n", pilot->start_pos, pilot->length0_min, pilot->length0_max, pilot->length1_min, pilot->length1_max );
            };
            return TRUE;
        };
    };
    mzload_pilot_reset_minmax_sync ( mzload );
    return FALSE;
}


static inline void mzload_pilot_search_end ( st_MZLOAD *mzload, uint32_t length0, uint32_t length1 ) {
    st_MZLOAD_PILOT *pilot = &mzload->pilot;
    if ( ( length0 >= ( pilot->length0_min - 1 ) ) && ( length0 <= ( pilot->length0_max + 1 ) ) ) {
        if ( ( length1 >= ( pilot->length1_min - 1 ) ) && ( length1 <= ( pilot->length1_max + 1 ) ) ) {
            return;
        };
    };
    pilot->have_pilot_end = TRUE;
    pilot->end_pos = mzload->pos - ( length0 + length1 );
    printf ( "Pilot end pos: %d, length0: %d, length1: %d\n", pilot->end_pos, length0, length1 );
}


static int mzload_read_pulse ( st_MZLOAD *mzload ) {

    en_CMT_STREAM_TYPE strtype = cmt_stream_get_stream_type ( mzload->stream );

    if ( CMT_STREAM_TYPE_BITSTREAM == strtype ) {
        st_CMT_BITSTREAM *bstr = mzload->stream->str.bitstream;
        uint32_t i;
        for ( i = mzload->pos; i < bstr->scans; i++ ) {
            int value = cmt_bitstream_get_value_on_position ( bstr, i );
            if ( value == 0 ) {
                mzload->loadTm = i - mzload->pos;
                mzload->bit_value = ( ( !mzload->sampTm ) || ( mzload->loadTm < mzload->sampTm ) ) ? 0 : 1;
                mzload->pos = i;
                for ( mzload->pos = i; mzload->pos < bstr->scans; mzload->pos++ ) {
                    int value = cmt_bitstream_get_value_on_position ( bstr, mzload->pos );
                    if ( value == 1 ) {
                        st_MZLOAD_PILOT *pilot = &mzload->pilot;
                        if ( !pilot->have_pilot_start ) {
                            uint32_t length0 = mzload->pos - i;
                            uint32_t length1 = mzload->loadTm;
                            if ( ( !pilot->minmax_synchronised ) || ( !mzload_pilot_search_start ( mzload, length0, length1 ) ) ) {
                                mzload_pilot_add_pulses ( mzload, length0, length0 );
                            };
                        } else if ( !pilot->have_pilot_end ) {
                            uint32_t length0 = mzload->pos - i;
                            uint32_t length1 = mzload->loadTm;
                            mzload_pilot_search_end ( mzload, length0, length1 );
                        };
                        return EXIT_SUCCESS;
                    };
                };
            };
        };
        // dosli jsme na konec streamu a nenasli jsme value = 1 !

    } else if ( CMT_STREAM_TYPE_VSTREAM == strtype ) {
        fprintf ( stderr, "%s():%d - Error: Unsupported stream type 'VSTREAM'\n", __func__, __LINE__ );
    } else {
        fprintf ( stderr, "%s():%d - Error: Unknown stream type '%d'\n", __func__, __LINE__, strtype );
    };

    mzload->status = MZLOAD_STATUS_EOF;

    return EXIT_FAILURE;
}


static int mzload_read_tmblock ( st_MZLOAD *mzload, int count, int rq_value ) {
    while ( count-- ) {
        if ( EXIT_FAILURE == mzload_read_pulse ( mzload ) ) return EXIT_FAILURE;
        if ( mzload->bit_value != rq_value ) {
            mzload->status = MZLOAD_STATUS_ERR_TMARK;
            return EXIT_SUCCESS;
        };
    };
    return EXIT_SUCCESS;
}


/**
 * Precteni symetrickeho tapemarku, ktery zacina blokem "1", nasledovanych stejnym blokem "0".
 * Za tapemarkem musi nasledovat jeden pulz s hodnotou "1".
 * 
 * @param mzload
 * @param tmark
 * @return 
 */
static int mzload_read_tapemark ( st_MZLOAD *mzload, en_MZLOAD_TAPEMARK tmark ) {

    // prvni bit z "1" bloku uz byl nacten jinde, takze ho odecteme
    if ( EXIT_FAILURE == mzload_read_tmblock ( mzload, ( tmark - 1 ), 1 ) ) return EXIT_FAILURE;
    if ( mzload->status != MZLOAD_STATUS_OK ) return EXIT_SUCCESS;

    if ( EXIT_FAILURE == mzload_read_tmblock ( mzload, tmark, 0 ) ) return EXIT_FAILURE;
    if ( mzload->status != MZLOAD_STATUS_OK ) return EXIT_SUCCESS;

    if ( EXIT_FAILURE == mzload_read_pulse ( mzload ) ) return EXIT_FAILURE;
    if ( mzload->bit_value != 1 ) {
        mzload->status = MZLOAD_STATUS_ERR_TMARK;
        return EXIT_SUCCESS;
    };

    printf ( "Tapemark - OK: %d\n", mzload->pos );

    return EXIT_SUCCESS;
}


int mzload_read_byte ( st_MZLOAD *mzload, uint8_t *byte, uint16_t *checksum ) {

    *byte = 0;

    if ( EXIT_FAILURE == mzload_read_pulse ( mzload ) ) return EXIT_FAILURE;

    if ( mzload->bit_value != 1 ) {
        mzload->status = MZLOAD_STATUS_ERR_DATA;
        return EXIT_FAILURE;
    };

    int i;

    for ( i = 0; i < 8; i++ ) {

        if ( EXIT_FAILURE == mzload_read_pulse ( mzload ) ) return EXIT_FAILURE;

        if ( ( checksum != NULL ) && ( mzload->bit_value ) ) *checksum = *checksum + 1;

        *byte <<= 1;
        *byte |= mzload->bit_value;
    };

    return EXIT_SUCCESS;
}


/**
 * Precteni cmt MZ blocku:
 * 
 * Pilot:
 * 
 * - nekolik tisic pulzu "0"
 * - tapemark => blok n x "1", n x "0"
 * - "1"
 * 
 * Data:
 * 
 * - "1"
 * - 8x hodnota bitu od nejvyssiho k nejnizsimu
 * 
 * Checksum (soucet vsech datovych jednicek):
 * 
 * - "1"
 * - 8x hodnota bitu od nejvyssiho k nejnizsimu
 * - "1"
 * - 8x hodnota bitu od nejvyssiho k nejnizsimu
 * 
 * @param mzload
 * @param tmark
 * @param size
 * @return 
 * 
 */
uint8_t* mzload_read_mzblock ( st_MZLOAD *mzload, en_MZLOAD_TAPEMARK tmark, uint16_t size ) {

    mzload_reset_pilot ( mzload );

    while ( 1 ) {

        mzload->status = MZLOAD_STATUS_OK;
        mzload->sampTm = 0;

        int i;
        for ( i = 0; i < MZLOAD_READ_SYNC; i++ ) {
            if ( EXIT_FAILURE == mzload_read_pulse ( mzload ) ) return NULL;
        };

        while ( mzload->status == MZLOAD_STATUS_OK ) {

            uint32_t totalScans = 0;

            int i;
            for ( i = 0; i < MZLOAD_SYNCHRO_BLOCK_SIZE; i++ ) {
                if ( EXIT_FAILURE == mzload_read_pulse ( mzload ) ) return NULL;

                st_MZLOAD_PILOT *pilot = &mzload->pilot;

                if ( ( pilot->have_pilot_end ) && ( mzload->bit_value == 1 ) ) {
                    if ( EXIT_FAILURE == mzload_read_tapemark ( mzload, tmark ) ) return NULL;
                    if ( mzload->status != MZLOAD_STATUS_OK ) {
                        break;
                    } else {

                        uint8_t *data = (uint8_t*) ui_utils_mem_alloc0 ( size );
                        uint16_t checksum = 0;

                        int i = 0;
                        while ( i < size ) {
                            if ( EXIT_FAILURE == mzload_read_byte ( mzload, &data[i++], &checksum ) ) {
                                g_free ( data );
                                return NULL;
                            };
                        };

                        uint8_t byte;
                        uint16_t ctrl_checksum = 0;

                        int ret = mzload_read_byte ( mzload, &byte, NULL );

                        if ( EXIT_FAILURE != ret ) {
                            ctrl_checksum = byte << 8;
                            ret = mzload_read_byte ( mzload, &byte, NULL );
                        };

                        if ( EXIT_FAILURE == ret ) {
                            g_free ( data );
                            return NULL;
                        };

                        ctrl_checksum |= byte;

                        if ( ctrl_checksum != checksum ) {
                            g_free ( data );
                            mzload->status = MZLOAD_STATUS_ERR_CHECKSUM;
                            return NULL;
                        };

                        return data;
                    };
                } else {
                    totalScans += mzload->loadTm;
                };
            };

            if ( mzload->status == MZLOAD_STATUS_OK ) {
                mzload->sampTm = totalScans / MZLOAD_SYNCHRO_BLOCK_DIVIDER;
            };
        };
    };

    return NULL;
}


/*
 * Fast IPL
 * type: 0xbb
 * size: 0x0000
 * strt: 0x1200
 * exec: 0x1110
 * 
 * komentar:
 * blocks, speed, size, start, exec
 * 
 * 
 */
int mzload_analyze_stream ( st_MZLOAD *mzload ) {

    st_MZF_HEADER *hdr = (st_MZF_HEADER*) mzload_read_mzblock ( mzload, MZLOAD_TAPEMARK_HEADER, sizeof ( st_MZF_HEADER ) );

    if ( !hdr ) {
        printf ( "Nejaka chyba '%d'!\n", mzload->status );
        return EXIT_FAILURE;
    };

#if 0
    uint8_t *data = (uint8_t*) hdr;

    FILE *fp = fopen ( "hdr.dat", "w" );
    fwrite ( data, 0x80, 1, fp );
    fclose ( fp );

    int i = 0;
    while ( i < 0x80 ) {
        int size = i + 8;
        for ( i = i; i < size; i++ ) {
            printf ( "0x%02x, ", data[i] );
            if ( i >= 0x80 ) break;
        };
        printf ( "\n" );
    };
#endif

    mzf_header_items_correction ( hdr );

    char ascii_filename[MZF_FNAME_FULL_LENGTH];
    mzf_tools_get_fname ( hdr, ascii_filename );

    printf ( "\nHeader found!\n" );
    printf ( "fname: %s\n", ascii_filename );
    printf ( "ftype: 0x%02x\n", hdr->ftype );
    printf ( "fstrt: 0x%04x\n", hdr->fstrt );
    printf ( "fsize: 0x%04x\n", hdr->fsize );
    printf ( "fexec: 0x%04x\n", hdr->fexec );

    uint8_t *body = NULL;

    if ( EXIT_SUCCESS == mzload_fastipl_test_loader ( hdr ) ) {

        printf ( "\nFast IPL v10.2 loader detect!\n" );

        uint8_t readpoint = mzload_fastipl_get_readpoint ( hdr );
        uint8_t blocks = mzload_fastipl_get_blcount ( hdr );

        hdr->fsize = mzload_fastipl_get_fsize ( hdr );
        hdr->fstrt = mzload_fastipl_get_fstrt ( hdr );
        hdr->fexec = mzload_fastipl_get_fexec ( hdr );

        printf ( "blocks: %d\n", blocks );
        printf ( "readpoint: 0x%02x\n", readpoint );
        printf ( "fstrt: 0x%04x\n", hdr->fstrt );
        printf ( "fsize: 0x%04x\n", hdr->fsize );
        printf ( "fexec: 0x%04x\n", hdr->fexec );

        int i = 0;
        while ( !body ) {
            i++;

            body = mzload_read_mzblock ( mzload, MZLOAD_TAPEMARK_BODY, hdr->fsize );

            if ( !body ) {
                if ( ( i < blocks ) && ( mzload->status != MZLOAD_STATUS_EOF ) ) {
                    fprintf ( stderr, "Error when reading file FastIPL body '%d'! Try next block...\n", mzload->status );
                } else {
                    fprintf ( stderr, "Error when reading file FastIPL body '%d'!\n", mzload->status );
                    g_free ( hdr );
                    return EXIT_FAILURE;
                };
            };
        };

        printf ( "\nFastIPL body - OK!\n" );

    } else {

        body = mzload_read_mzblock ( mzload, MZLOAD_TAPEMARK_BODY, hdr->fsize );

        if ( !body ) {
            fprintf ( stderr, "Error when reading file body '%d'!\n", mzload->status );
            g_free ( hdr );
            return EXIT_FAILURE;
        };

        printf ( "\nBody - OK!\n" );

#if 0
        FILE *fp = fopen ( "body.dat", "w" );
        fwrite ( body, hdr->fsize, 1, fp );
        fclose ( fp );

        int i = 0;
        while ( i < hdr->fsize ) {
            int size = i + 8;
            for ( i = i; i < size; i++ ) {
                printf ( "0x%02x, ", body[i] );
                if ( i >= hdr->fsize ) break;
            };
            printf ( "\n" );
        };
#endif

        if ( EXIT_SUCCESS == mzload_turbo_test_loader ( hdr, body ) ) {
            printf ( "\nTurboCopy v1.21 loader detect!\n" );

            uint8_t readpoint = mzload_turbo_get_readpoint ( body );
            uint8_t blocks = mzload_turbo_get_blcount ( body );

            hdr->fsize = mzload_turbo_get_fsize ( body );
            hdr->fstrt = mzload_turbo_get_fstrt ( body );
            hdr->fexec = mzload_turbo_get_fexec ( body );

            g_free ( body );
            body = NULL;

            printf ( "blocks: %d\n", blocks );
            printf ( "readpoint: 0x%02x\n", readpoint );
            printf ( "fstrt: 0x%04x\n", hdr->fstrt );
            printf ( "fsize: 0x%04x\n", hdr->fsize );
            printf ( "fexec: 0x%04x\n", hdr->fexec );

            int i = 0;
            while ( !body ) {
                i++;

                body = mzload_read_mzblock ( mzload, MZLOAD_TAPEMARK_BODY, hdr->fsize );

                if ( !body ) {
                    if ( ( i < blocks ) && ( mzload->status != MZLOAD_STATUS_EOF ) ) {
                        fprintf ( stderr, "Error when reading file TURBO body '%d'! Try next block...\n", mzload->status );
                    } else {
                        fprintf ( stderr, "Error when reading file TURBO body '%d'!\n", mzload->status );
                        g_free ( hdr );
                        return EXIT_FAILURE;
                    };
                };
            };

            printf ( "\nTURBO body - OK!\n" );
        };
    };

    g_free ( hdr );
    g_free ( body );

    return EXIT_SUCCESS;
}


void mzload_test ( void ) {

    //char *filename = "CMT/WAV/cisty_rom_z_emu.wav";

    //char *filename = "CMT/WAV/tc_loader_2_1.wav";
    //char *filename = "CMT/WAV/tc_loader_2_1_tc.wav";
    //char *filename = "CMT/WAV/tc_loader_2_1_tc3.wav";
    //char *filename = "CMT/WAV/ic_loader_2_1_tc3b.wav";

    //char *filename = "CMT/WAV/save_Intercopy_S1400_1381.wav";
    //char *filename = "CMT/WAV/save_Intercopy_S1800_1815.wav";
    //char *filename = "CMT/WAV/save_Intercopy_S2100_2116.wav";
    //char *filename = "CMT/WAV/save_Intercopy_S2500_2535.wav";

    //char *filename = "CMT/WAV/save_ROM_1152.wav";
    //char *filename = "CMT/WAV/save_Intercopy_N1200_1149.wav";
    //char *filename = "CMT/WAV/save_Intercopy_N2400_2262.wav";
    //char *filename = "CMT/WAV/save_Intercopy_N2800_2701.wav";
    char *filename = "CMT/WAV/save_Intercopy_N3200_2945.wav";


    printf ( "\n\n\nMZLoad analyzer open: %s\n", filename );

    st_MZLOAD *mzload = mzload_create ( filename );
    if ( !mzload ) {
        return;
    };

    mzload_analyze_stream ( mzload );

    mzload_destroy ( mzload );
}
