/* 
 * File:   main.c
 * Author: Michal Hucik <hucik@ordoz.com>
 *
 * Created on 27. září 2018, 20:03
 * 
 * 
 * ----------------------------- License -------------------------------------
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * ---------------------------------------------------------------------------
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <glib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <locale.h>

#include "ui/generic_driver/ui_memory_driver.h"
#include "libs/cmt_stream/cmt_stream_all.h"
#include "libs/cmttool/cmttool.h"
#include "libs/mzf/mzf_tools.h"
#include "ui/ui_utils.h"

static st_DRIVER *g_driver = &g_ui_memory_driver_realloc;


typedef struct st_MZCMTTOOL_MAIN {
    GOptionContext *opt_context;
    gchar **opt_args;
    gchar *input_filename;
    gboolean verbose;
    gboolean inverted;
    gboolean forced_overwrite;
    gdouble start_time;
    guint32 start_pos;
    gchar *destdir;
    gint32 save_id;
    gchar *output_filename;
    st_CMTTOOL *cmttool;
} st_MZCMTTOOL_MAIN;

static st_MZCMTTOOL_MAIN *g_mzcmt = NULL;


static void mzcmttool_print_files ( st_CMTTOOL *cmttool ) {

    st_CMT_STREAM *stream = cmttool_get_stream ( cmttool );
    gdouble scan_time = cmt_stream_get_scantime ( stream );

    const gchar *fname_spaces = "                 ";
    gchar ascii_filename[MZF_FNAME_FULL_LENGTH];
    gchar ascii_filename_full[MZF_FNAME_FULL_LENGTH];
    snprintf ( ascii_filename_full, MZF_FNAME_FULL_LENGTH, "File name%s", fname_spaces );

    const gchar *fparam_spaces = "                     ";
    gchar fparam[21];

    g_print ( "\nID\tFormat\tType %s\tSize   Start  Exec   Pos.(s)\n", ascii_filename_full );

    guint32 count_files = cmttool_get_count_files ( cmttool );
    guint32 i;
    for ( i = 0; i < count_files; i++ ) {

        st_CMTTOOL_FILE *cf = cmttool_get_file ( cmttool, i );
        st_MZF_HEADER *hdr = cmttool_file_get_header ( cf );

        mzf_tools_get_fname ( hdr, ascii_filename );
        snprintf ( ascii_filename_full, MZF_FNAME_FULL_LENGTH, "%s%s", ascii_filename, fname_spaces );
        guint32 start_pos = cmttool_file_get_start_pos ( cf );

        en_CMTTOOL_FORMAT format = cmttool_file_get_format ( cf );

        if ( format != CMTTOOL_FORMAT_BSD ) {
            snprintf ( fparam, sizeof ( fparam ), "0x%04x 0x%04x 0x%04x", hdr->fsize, hdr->fstrt, hdr->fexec );
        } else {
            snprintf ( fparam, sizeof ( fparam ), "0x%04x%s", cmttool_file_get_body_size ( cf ), fparam_spaces );
        };

        g_print ( "%d\t%s\t0x%02x %s\t%s %f\n",
                  i,
                  cmttool_file_get_format_txt ( cf ),
                  hdr->ftype,
                  ascii_filename_full,
                  fparam,
                  (gdouble) start_pos * scan_time
                  );
    };
}


static int mzcmttool_save_file ( st_CMTTOOL_FILE *cf, const gchar *directory, const gchar *filename, gboolean forced ) {


    typedef enum en_MZCMTTOOL_SAVE_TYPE {
        MZCMTTOOL_SAVE_TYPE_DAT = 0,
        MZCMTTOOL_SAVE_TYPE_MZF,
    } en_MZCMTTOOL_SAVE_TYPE;

    if ( directory ) {

        if ( !g_file_test ( directory, G_FILE_TEST_EXISTS ) ) {
            printf ( "Creating directory '%s'\n", directory );
            if ( -1 == g_mkdir_with_parents ( directory, 0775 ) ) {
                fprintf ( stderr, "%s():%d - Can't create directory '%s'\n\n", __func__, __LINE__, directory );
                return EXIT_FAILURE;
            };
        };

        if ( !g_file_test ( directory, G_FILE_TEST_IS_DIR ) ) {
            fprintf ( stderr, "%s():%d - This is not a directory '%s'\n\n", __func__, __LINE__, directory );
            return EXIT_FAILURE;
        };
    };

    gchar ascii_filename[MZF_FNAME_FULL_LENGTH];
    st_MZF_HEADER *hdr = cmttool_file_get_header ( cf );
    mzf_tools_get_fname ( hdr, ascii_filename );

    gchar *save_filename = NULL;
    en_MZCMTTOOL_SAVE_TYPE save_type = ( cmttool_file_get_format ( cf ) == CMTTOOL_FORMAT_BSD ) ? MZCMTTOOL_SAVE_TYPE_DAT : MZCMTTOOL_SAVE_TYPE_MZF;

    if ( filename ) {
        if ( directory ) {
            save_filename = g_build_filename ( directory, filename, NULL );
        } else {
            save_filename = g_strdup ( filename );
        };

        gboolean have_extension = FALSE;
        gchar *ext = g_strrstr ( filename, "." );
        if ( ext ) {
            ext++;
            if ( 0 == strcasecmp ( ext, "mzf" ) ) {
                have_extension = TRUE;
                save_type = MZCMTTOOL_SAVE_TYPE_MZF;
            } else if ( 0 == strcasecmp ( ext, "dat" ) ) {
                have_extension = TRUE;
                save_type = MZCMTTOOL_SAVE_TYPE_DAT;
            };
        };

        if ( !have_extension ) {
            GString *str = g_string_new ( save_filename );
            g_free ( save_filename );
            g_string_append_printf ( str, ".%s", ( save_type == MZCMTTOOL_SAVE_TYPE_DAT ) ? "dat" : "mzf" );
            save_filename = g_string_free ( str, FALSE );
        };

        if ( g_file_test ( save_filename, G_FILE_TEST_EXISTS ) && !forced ) {
            g_print ( "Error - ouput file already exists '%s'\n\n", save_filename );
            g_free ( save_filename );
            return EXIT_FAILURE;
        };

    } else {

        guint instance = 0;

        do {

            if ( directory ) {
                save_filename = g_build_filename ( directory, ascii_filename, NULL );
            } else {
                save_filename = g_strdup ( ascii_filename );
            };

            GString *str = g_string_new ( save_filename );
            g_free ( save_filename );
            if ( instance ) g_string_append_printf ( str, "_%d", instance );
            instance++;
            g_string_append_printf ( str, ".%s", ( save_type == MZCMTTOOL_SAVE_TYPE_DAT ) ? "dat" : "mzf" );
            save_filename = g_string_free ( str, FALSE );

            if ( g_file_test ( save_filename, G_FILE_TEST_EXISTS ) && !forced ) {
                g_free ( save_filename );
                save_filename = NULL;
            };

        } while ( save_filename == NULL );

    };

    g_print ( "Saving file ID %d '%s' into: %s\n",
              cmttool_file_get_id ( cf ),
              ascii_filename,
              save_filename
              );

    st_HANDLER *handler = generic_driver_open_memory ( NULL, g_driver, 1 );

    int ret = EXIT_FAILURE;

    if ( !handler ) {
        fprintf ( stderr, "%s():%d - Could not open handler\n", __func__, __LINE__ );
    } else {

        handler->spec.memspec.swelling_enabled = 1;

        if ( save_type == MZCMTTOOL_SAVE_TYPE_MZF ) {

            if ( EXIT_SUCCESS != mzf_write_header ( handler, hdr ) ) {
                fprintf ( stderr, "%s():%d - Could not write MZF header\n", __func__, __LINE__ );
            } else {
                ret = mzf_write_body ( handler, cmttool_file_get_body ( cf ), cmttool_file_get_body_size ( cf ) );
            };

        } else {
            ret = mzf_write_body_on_offset ( handler, 0, cmttool_file_get_body ( cf ), cmttool_file_get_body_size ( cf ) );
        };
    };

    if ( ret == EXIT_SUCCESS ) {
        ret = generic_driver_save_memory ( handler, save_filename );
        if ( EXIT_SUCCESS != generic_driver_save_memory ( handler, save_filename ) ) {
            fprintf ( stderr, "%s():%d - Could not open file '%s': %s\n", __func__, __LINE__, save_filename, strerror ( errno ) );
        };
    };

    generic_driver_close ( handler );
    if ( handler ) g_free ( handler );
    g_free ( save_filename );

    return ret;
}


static int mzcmttool_work ( st_MZCMTTOOL_MAIN *mzcmt ) {

    guint32 start_offset = 0;

    en_CMT_STREAM_POLARITY polarity = ( !mzcmt->inverted ) ? CMT_STREAM_POLARITY_NORMAL : CMT_STREAM_POLARITY_INVERTED;

    ui_memory_driver_init ( );

    g_print ( "Analyzer open file: %s\n", mzcmt->input_filename );
    g_print ( "Polarity: %s\n", ( !mzcmt->inverted ) ? "normal" : "inverted" );

    mzcmt->cmttool = cmttool_create_from_wav ( mzcmt->input_filename, polarity );
    if ( !mzcmt->cmttool ) {
        return EXIT_FAILURE;
    };

    st_CMT_STREAM *stream = cmttool_get_stream ( mzcmt->cmttool );
    gdouble scan_time = cmt_stream_get_scantime ( stream );

    if ( mzcmt->start_time != 0 ) {
        start_offset = mzcmt->start_time / scan_time;
    } else if ( mzcmt->start_pos != 0 ) {
        start_offset = mzcmt->start_pos;
    };

    g_print ( "Start time: %f s\n\n", (gdouble) start_offset * scan_time );

    cmttool_set_debug ( mzcmt->cmttool, mzcmt->verbose );
    guint count_files = cmttool_analyze ( mzcmt->cmttool, start_offset );
    g_print ( "Found %d file(s)\n", count_files );

    int ret = EXIT_SUCCESS;

    if ( count_files ) {
        if ( mzcmt->destdir != NULL ) {
            g_print ( "Saving all files into: %s\n", mzcmt->destdir );
            guint i;
            for ( i = 0; i < count_files; i++ ) {
                st_CMTTOOL_FILE *cf = cmttool_get_file ( mzcmt->cmttool, i );
                ret = mzcmttool_save_file ( cf, mzcmt->destdir, NULL, mzcmt->forced_overwrite );
                if ( ret != EXIT_SUCCESS ) break;
            };

        } else if ( mzcmt->save_id != -1 ) {
            if ( count_files > mzcmt->save_id ) {
                st_CMTTOOL_FILE *cf = cmttool_get_file ( mzcmt->cmttool, mzcmt->save_id );
                ret = mzcmttool_save_file ( cf, NULL, mzcmt->output_filename, mzcmt->forced_overwrite );
            } else {
                g_print ( "Error - file ID %d not found\n", mzcmt->save_id );
                ret = EXIT_FAILURE;
            };
        } else {
            mzcmttool_print_files ( mzcmt->cmttool );
        };

        g_print ( "\n" );
    };


    return ret;
}


void main_app_quit ( int retcode ) {
    if ( g_mzcmt ) {
        if ( g_mzcmt->opt_context ) g_option_context_free ( g_mzcmt->opt_context );
        if ( g_mzcmt->opt_args ) g_strfreev ( g_mzcmt->opt_args );
        if ( g_mzcmt->input_filename ) g_free ( g_mzcmt->input_filename );
        if ( g_mzcmt->output_filename ) g_free ( g_mzcmt->output_filename );
        if ( g_mzcmt->cmttool ) cmttool_destroy ( g_mzcmt->cmttool );
        g_free ( g_mzcmt );
    };
    exit ( retcode );
}


/*
 * TODO: vyzkouset diakritiku
 * TODO: joe loader?
 */
int main ( int argc, char** argv ) {

    setlocale ( LC_ALL, "" );

    GError *error = NULL;

    g_mzcmt = (st_MZCMTTOOL_MAIN*) ui_utils_mem_alloc0 ( sizeof ( st_MZCMTTOOL_MAIN ) );
    g_mzcmt->save_id = -1;

    GOptionEntry entries[] = {
        { "save", 's', 0, G_OPTION_ARG_INT, &g_mzcmt->save_id, "Save selected file into mzf|dat\n\t\t\t(output filename is not necessary)", "ID" },
        { "destdir", 'd', 0, G_OPTION_ARG_FILENAME, &g_mzcmt->destdir, "Save all files into destination dir", "dir" },
        { "forced", 'f', 0, G_OPTION_ARG_NONE, &g_mzcmt->forced_overwrite, "Forced overwrite mode", NULL },
        { "time", 't', 0, G_OPTION_ARG_DOUBLE, &g_mzcmt->start_time, "Start time offset of iput file", "0.000" },
        { "position", 'p', 0, G_OPTION_ARG_INT, &g_mzcmt->start_pos, "Start sample position offset of iput file", "0" },
        { "inverted", 'i', 0, G_OPTION_ARG_NONE, &g_mzcmt->inverted, "Inverted CMT stream polarity", NULL },
        { "verbose", 'v', 0, G_OPTION_ARG_NONE, &g_mzcmt->verbose, "Verbose mode", NULL },
        { NULL }
    };

    const gchar *options_comment = ""
            "This utility analyzes the Sharp MZ CMT stream in WAV audio format.\n\n"
            "Supported MZ CMT file logic formats:\n\n"
            "\t- NORMAL\n"
            "\t- BSD\n"
            "\t- TURBO (ver: 1, 1.2x)\n"
            "\t- FastIPL (ver: 2, 7 - 10.2)";

    g_print ( "MZ cmttool ver 1.0\n\n" );

    //#ifdef G_OS_WIN32
#ifdef WINDOWS
    g_mzcmt->opt_args = g_win32_get_command_line ( );
#else
    g_mzcmt->opt_args = g_strdupv ( argv );
#endif

    g_mzcmt->opt_context = g_option_context_new ( "<input.wav> [output.[mzf|dat]]" );
    g_option_context_set_summary ( g_mzcmt->opt_context, options_comment );
    g_option_context_set_description ( g_mzcmt->opt_context, "Programmed by: Michal Hučík - ORDOZ <hucik@ordoz.com>\n" );
    g_option_context_add_main_entries ( g_mzcmt->opt_context, entries, NULL );

    if ( !g_option_context_parse_strv ( g_mzcmt->opt_context, &g_mzcmt->opt_args, &error ) ) {
        g_print ( "Error - option parsing failed: %s\n\n", error->message );
        g_print ( g_option_context_get_help ( g_mzcmt->opt_context, TRUE, NULL ) );
        main_app_quit ( EXIT_FAILURE );
    };

    if ( ( g_mzcmt->destdir != NULL ) && ( g_mzcmt->save_id != -1 ) ) {
        g_print ( "Error - invalid combination of -s and -d parametters!\n\n" );
        g_print ( g_option_context_get_help ( g_mzcmt->opt_context, TRUE, NULL ) );
        main_app_quit ( EXIT_FAILURE );
    };

    if ( ( g_mzcmt->start_time != 0 ) && ( g_mzcmt->start_pos != 0 ) ) {
        g_print ( "Error - invalid combination of -t and -p parametters!\n\n" );
        g_print ( g_option_context_get_help ( g_mzcmt->opt_context, TRUE, NULL ) );
        main_app_quit ( EXIT_FAILURE );
    };

    if ( g_strv_length ( g_mzcmt->opt_args ) < 2 ) {
        g_print ( "Error - input WAV file parameter mising!\n\n" );
        g_print ( g_option_context_get_help ( g_mzcmt->opt_context, TRUE, NULL ) );
        main_app_quit ( EXIT_FAILURE );
    };

    g_mzcmt->input_filename = g_strdup ( g_mzcmt->opt_args[1] );

    if ( ( g_mzcmt->save_id != -1 ) && ( g_strv_length ( g_mzcmt->opt_args ) >= 3 ) ) {
        g_mzcmt->output_filename = g_strdup ( g_mzcmt->opt_args[2] );
    };

    g_option_context_free ( g_mzcmt->opt_context );
    g_mzcmt->opt_context = NULL;
    g_strfreev ( g_mzcmt->opt_args );
    g_mzcmt->opt_args = NULL;

    main_app_quit ( mzcmttool_work ( g_mzcmt ) );
    return EXIT_SUCCESS;
}

